# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import shutil
import subprocess
import sys
from pathlib import Path

from sphinx.ext.autodoc import Documenter, MethodDocumenter

import woodwork

path = os.path.join("..", "..")
sys.path.insert(0, os.path.abspath(path))


# -- Project information -----------------------------------------------------

project = "Woodwork"
copyright = "2020, Alteryx, Inc."
author = "Alteryx, Inc."

# The short X.Y version
version = woodwork.__version__
# The full version, including alpha/beta/rc tags
release = woodwork.__version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.extlinks",
    "sphinx.ext.inheritance_diagram",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
    "sphinx_inline_tabs",
    "sphinx_copybutton",
    "myst_parser",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The main toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["**.ipynb_checkpoints"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "pygment_light_style": "tango",
    "pygment_dark_style": "native",
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/alteryx/woodwork",
            "icon": "fab fa-github-square",
            "type": "fontawesome",
        },
        {
            "name": "Twitter",
            "url": "https://twitter.com/AlteryxOSS",
            "icon": "fab fa-twitter-square",
            "type": "fontawesome",
        },
        {
            "name": "Slack",
            "url": "https://join.slack.com/t/alteryx-oss/shared_invite/zt-182tyvuxv-NzIn6eiCEf8TBziuKp0bNA",
            "icon": "fab fa-slack",
            "type": "fontawesome",
        },
    ],
    "collapse_navigation": False,
    "navigation_depth": 2,
}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "images/woodwork.svg"

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "images/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "Woodworkdoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "Woodwork.tex",
        "Woodwork Documentation",
        "Alteryx, Inc.",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "woodwork",
        "Woodwork Documentation",
        [author],
        1,
    ),
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "Woodwork",
        "Woodwork Documentation",
        author,
        "Woodwork",
        "One line description of project.",
        "Miscellaneous",
    ),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]

# -- Options for Markdown files ----------------------------------------------

myst_admonition_enable = True
myst_deflist_enable = True
myst_heading_anchors = 3

# -- Options for Sphinx Copy Button ------------------------------------------

copybutton_prompt_text = "myinputprompt"
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- Extension configuration -------------------------------------------------

# If woodwork is open-sourced: replace github specific style.css
extlinks = {
    "issue": ("https://github.com/alteryx/woodwork/issues/%s", "GH#%s"),
    "pr": ("https://github.com/alteryx/woodwork/pull/%s", "GH#%s"),
    "user": ("https://github.com/%s", "@%s"),
}

autosummary_generate = ["api_reference.rst"]
templates_path = ["_templates"]

html_show_sphinx = False
nbsphinx_execute = "always"
nbsphinx_timeout = (
    600  # sphinx defaults each cell to 30 seconds so we need to override here
)

inheritance_graph_attrs = dict(
    rankdir="LR",
    size='"1000, 333"',
    fontsize=30,
    labelfontsize=30,
    ratio="compress",
    dpi=960,
)


class AccessorLevelDocumenter(Documenter):
    """
    Documenter subclass for objects on accessor level (methods, attributes).

    Referenced pandas-sphinx-theme (https://github.com/pandas-dev/pandas-sphinx-theme)
    and sphinx-doc (https://github.com/sphinx-doc/sphinx/blob/8c7faed6fcbc6b7d40f497698cb80fc10aee1ab3/sphinx/ext/autodoc/__init__.py#L846)
    """

    def resolve_name(self, modname, parents, path, base):
        modname = "woodwork"
        mod_cls = path.rstrip(".")
        mod_cls = mod_cls.split(".")
        return modname, mod_cls + [base]


class AccessorCallableDocumenter(AccessorLevelDocumenter, MethodDocumenter):
    """
    This documenter lets us removes .__call__ from the method signature for
    callable accessors like Series.plot
    """

    objtype = "accessorcallable"
    directivetype = "method"

    # lower than MethodDocumenter; otherwise the doc build prints warnings
    priority = 0.5

    def format_name(self):
        return MethodDocumenter.format_name(self).rstrip(".__call__")


class AccessorMethodDocumenter(AccessorLevelDocumenter, MethodDocumenter):
    objtype = "accessormethod"
    directivetype = "method"

    # lower than MethodDocumenter so this is not chosen for normal methods
    priority = 0.6


def setup(app):
    home_dir = os.environ.get("HOME", "/")
    ipython_p = Path(home_dir + "/.ipython/profile_default/startup")
    ipython_p.mkdir(parents=True, exist_ok=True)
    file_p = os.path.abspath(os.path.dirname(__file__))
    shutil.copy(
        file_p + "/set-headers.py",
        home_dir + "/.ipython/profile_default/startup",
    )
    app.add_js_file(
        "https://cdnjs.cloudflare.com/ajax/libs/require.js/2.1.10/require.min.js",
    )
    app.add_css_file("style.css")
    app.add_autodocumenter(AccessorCallableDocumenter)
    app.add_autodocumenter(AccessorMethodDocumenter)



from importlib import resources as pkg_resources
from typing import Any, Callable, Iterable, Union

import numpy as np
import pandas as pd
from pandas.api import types as pdtypes

import woodwork as ww
from woodwork import data
from woodwork.type_sys.utils import _is_categorical_series, col_is_datetime

Tokens = Iterable[str]

COMMON_WORDS_SET = set(
    line.strip().lower() for line in pkg_resources.open_text(data, "1-1000.txt")
)

NL_delimiters = r"[- \[\].,!\?;\n]"


def categorical_func(series):
    if pdtypes.is_categorical_dtype(series.dtype):
        return True

    if pdtypes.is_string_dtype(series.dtype) and not col_is_datetime(series):
        categorical_threshold = ww.config.get_option("categorical_threshold")

        return _is_categorical_series(series, categorical_threshold)

    if pdtypes.is_float_dtype(series.dtype) or pdtypes.is_integer_dtype(series.dtype):
        numeric_categorical_threshold = ww.config.get_option(
            "numeric_categorical_threshold",
        )
        if numeric_categorical_threshold is not None:
            return _is_categorical_series(series, numeric_categorical_threshold)
        else:
            return False

    return False


def integer_func(series):
    if integer_nullable_func(series) and not series.isnull().any():
        return all(series.mod(1).eq(0))
    return False


def integer_nullable_func(series):
    if pdtypes.is_integer_dtype(series.dtype):
        threshold = ww.config.get_option("numeric_categorical_threshold")
        if threshold is not None:
            return not _is_categorical_series(series, threshold)
        else:
            return True
    elif pdtypes.is_float_dtype(series.dtype):
        if not series.isnull().any():
            return False
        series_no_null = series.dropna()
        return all(series_no_null.mod(1).eq(0))

    return False


def double_func(series):
    if pdtypes.is_float_dtype(series.dtype):
        threshold = ww.config.get_option("numeric_categorical_threshold")
        if threshold is not None:
            return not _is_categorical_series(series, threshold)
        else:
            return True

    return False


def boolean_func(series):
    if boolean_nullable_func(series) and not series.isnull().any():
        return True
    return False


def boolean_nullable_func(series):
    if pdtypes.is_bool_dtype(series.dtype) and not pdtypes.is_categorical_dtype(
        series.dtype,
    ):
        return True
    elif pdtypes.is_object_dtype(series.dtype):
        series_no_null = series.dropna()
        try:
            series_no_null_unq = set(series_no_null)
            if series_no_null_unq in [
                {False, True},
                {True},
                {False},
            ]:
                return True
        except TypeError:  # Necessary to check for non-hashable values because of object dtype consideration
            return False
    return False


def datetime_func(series):
    if col_is_datetime(series):
        return True
    return False


def timedelta_func(series):
    if pdtypes.is_timedelta64_dtype(series.dtype):
        return True
    return False


def num_common_words(wordlist: Union[Tokens, Any]) -> float:
    if not isinstance(wordlist, Iterable):
        return np.nan
    num_common_words = 0
    for x in wordlist:
        if x.lower() in COMMON_WORDS_SET:
            num_common_words += 1
    return num_common_words


def natural_language_func(series):
    tokens = series.astype("string").str.split(NL_delimiters)
    mean_num_common_words = np.nanmean(tokens.map(num_common_words))

    return (
        mean_num_common_words > 1.14
    )  # determined through https://github.com/alteryx/nl_inference


class InferWithRegex:
    def __init__(self, get_regex: Callable[[], str]):
        self.get_regex = get_regex

    def __call__(self, series: pd.Series) -> bool:
        series = series.dropna()
        regex = self.get_regex()

        # Includes a check for object dtypes
        if not pdtypes.is_string_dtype(series.dtype):
            return False

        try:
            series_match_method = series.str.match
        except (AttributeError, TypeError):
            # This can happen either when the inferred dtype for a series is not
            # compatible with the pandas string API (AttributeError) *or* when the
            # inferred dtype is not compatible with the string API `match` method
            # (TypeError)
            return False
        matches = series_match_method(pat=regex)

        return matches.sum() == len(matches)


email_address_func = InferWithRegex(
    lambda: ww.config.get_option("email_inference_regex"),
)
phone_number_func = InferWithRegex(
    lambda: ww.config.get_option("phone_inference_regex"),
)
postal_code_func = InferWithRegex(
    lambda: ww.config.get_option("postal_code_inference_regex"),
)
url_func = InferWithRegex(lambda: ww.config.get_option("url_inference_regex"))
ip_address_func = InferWithRegex(
    lambda: (
        "("
        + ww.config.get_option("ipv4_inference_regex")
        + "|"
        + ww.config.get_option("ipv6_inference_regex")
        + ")"
    ),
)



import re
from datetime import datetime

import numpy as np
import pandas as pd
import pytest

from woodwork.accessor_utils import _is_dask_series, _is_spark_series, init_series
from woodwork.exceptions import (
    TypeConversionError,
    TypeConversionWarning,
    TypeValidationError,
)
from woodwork.logical_types import (
    URL,
    Age,
    AgeFractional,
    AgeNullable,
    Boolean,
    BooleanNullable,
    Categorical,
    Datetime,
    Double,
    EmailAddress,
    Integer,
    IntegerNullable,
    LatLong,
    Ordinal,
    PhoneNumber,
    PostalCode,
    _replace_nans,
)
from woodwork.tests.testing_utils.table_utils import (
    concat_dataframe_or_series,
    to_pandas,
)
from woodwork.utils import import_or_none

ps = import_or_none("pyspark.pandas")


def test_logical_eq():
    assert Boolean == Boolean
    assert Boolean() == Boolean()
    assert Categorical != Boolean
    assert Datetime != Datetime()
    assert Datetime() == Datetime(datetime_format=None)
    assert Datetime() != Datetime(datetime_format="%Y-%m-%d")


def test_logical_repr():
    assert isinstance(repr(Boolean), str)
    assert repr(Boolean) == "Boolean"
    assert isinstance(repr(Categorical), str)
    assert repr(Categorical) == "Categorical"


def test_instantiated_type_str():
    assert str(Categorical()) == "Categorical"
    assert str(Boolean()) == "Boolean"


def test_ordinal_order_errors():
    series = pd.Series([1, 2, 3]).astype("category")

    with pytest.raises(
        TypeError,
        match="Order values must be specified in a list or tuple",
    ):
        Ordinal(order="not_valid").transform(series)

    with pytest.raises(ValueError, match="Order values cannot contain duplicates"):
        Ordinal(order=["a", "b", "b"]).transform(series)


def test_ordinal_init_with_order():
    order = ["bronze", "silver", "gold"]
    ordinal_from_list = Ordinal(order=order)
    assert ordinal_from_list.order == order
    assert str(ordinal_from_list) == "Ordinal: ['bronze', 'silver', 'gold']"

    order = ("bronze", "silver", "gold")
    ordinal_from_tuple = Ordinal(order=order)
    assert ordinal_from_tuple.order == order
    assert str(ordinal_from_list) == "Ordinal: ['bronze', 'silver', 'gold']"


def test_ordinal_transform_validates(ordinal_transform_series_pandas) -> None:
    typ = Ordinal(order=None)
    with pytest.raises(TypeError, match=r"order values defined"):
        typ.transform(ordinal_transform_series_pandas)


def test_ordinal_transform_pandas(ordinal_transform_series_pandas) -> None:
    order = [2, 1, 3]
    typ = Ordinal(order=order)
    ser_ = typ.transform(ordinal_transform_series_pandas)

    assert ser_.dtype == "category"
    pd.testing.assert_index_equal(ser_.cat.categories, pd.Int64Index(order))


def test_ordinal_transform_dask(ordinal_transform_series_dask) -> None:
    order = [2, 1, 3]
    typ = Ordinal(order=order)
    ser_ = typ.transform(ordinal_transform_series_dask).compute()

    assert ser_.dtype == "category"
    pd.testing.assert_index_equal(ser_.cat.categories, pd.Int64Index(order))


def test_ordinal_transform_spark(ordinal_transform_series_spark) -> None:
    order = [2, 1, 3]
    typ = Ordinal(order=order)
    ser_ = typ.transform(ordinal_transform_series_spark)

    assert ser_.dtype == pd.StringDtype()


def test_get_valid_dtype(sample_series):
    valid_dtype = Categorical._get_valid_dtype(type(sample_series))
    if _is_spark_series(sample_series):
        assert valid_dtype == "string"
    else:
        assert valid_dtype == "category"

    valid_dtype = Boolean._get_valid_dtype(type(sample_series))
    assert valid_dtype == "bool"


def test_latlong_transform(latlong_df):
    df_type = str(type(latlong_df))
    dask = "dask" in df_type
    spark = "spark" in df_type
    nan = float("nan")

    expected_data = {
        "tuple_ints": [(1.0, 2.0), (3.0, 4.0)],
        "tuple_strings": [(1.0, 2.0), (3.0, 4.0)],
        "string_tuple": [(1.0, 2.0), (3.0, 4.0)],
        "bracketless_string_tuple": [(1.0, 2.0), (3.0, 4.0)],
        "list_strings": [(1.0, 2.0), (3.0, 4.0)],
        "combo_tuple_types": [(1.0, 2.0), (3.0, 4.0)],
        "null_value": [nan, (3.0, 4.0)],
        "null_latitude": [(nan, 2.0), (3.0, 4.0)],
        "both_null": [(nan, nan), (3.0, 4.0)],
    }

    latlong = LatLong()
    for column in latlong_df:
        series = latlong_df[column]
        actual = latlong.transform(series)

        if dask:
            actual = actual.compute()
        elif spark:
            actual = actual.to_pandas()

        expected = pd.Series(expected_data[column], name=column)
        pd.testing.assert_series_equal(actual, expected)


def test_latlong_transform_empty_series(empty_latlong_df):
    latlong = LatLong()
    series = empty_latlong_df["latlong"]
    actual = latlong.transform(series)

    if _is_dask_series(actual):
        actual = actual.compute()
    elif _is_spark_series(actual):
        actual = actual.to_pandas()

    assert actual.empty
    assert actual.name == "latlong"
    assert actual.dtype == latlong.primary_dtype


def test_latlong_validate(latlong_df):
    error_message = re.escape(
        "Cannot initialize Woodwork. Series does not contain properly formatted "
        "LatLong data. Try reformatting before initializing or use the "
        "woodwork.init_series function to initialize.",
    )
    latlong = LatLong()
    series = latlong_df["tuple_ints"]
    new_series = init_series(series, logical_type=LatLong)
    latlong.validate(new_series)
    with pytest.raises(TypeValidationError, match=error_message):
        latlong.validate(series)


def test_datetime_transform(datetimes):
    datetime = Datetime()
    for series in datetimes:
        assert str(series.dtype) == "object"
        transform = datetime.transform(series)
        assert str(transform.dtype) == "datetime64[ns]"
        assert datetime.datetime_format is not None


def test_datetime_inference_ambiguous_format():
    datetime = Datetime()
    dates = pd.Series(["01/01/2017"] * 2 + ["13/12/2017"], name="dates")
    warning = (
        "Some rows in series 'dates' are incompatible with datetime format "
        "'%m/%d/%Y' and have been replaced with null values. You may be able "
        "to fix this by using an instantiated Datetime logical type with a different "
        "format string specified for this column during Woodwork initialization."
    )
    with pytest.warns(TypeConversionWarning, match=warning):
        transformed = datetime.transform(dates)
    assert str(transformed.dtype) == "datetime64[ns]"
    assert transformed[2] is pd.NaT
    assert datetime.datetime_format == "%m/%d/%Y"


def test_datetime_coerce_user_format():
    datetime = Datetime(datetime_format="%m/%d/%Y")
    dates = pd.Series(["01/01/2017"] * 2 + ["13/12/2017"], name="dates")
    warning = (
        "Some rows in series 'dates' are incompatible with datetime format "
        "'%m/%d/%Y' and have been replaced with null values. You may be able "
        "to fix this by using an instantiated Datetime logical type with a different "
        "format string specified for this column during Woodwork initialization."
    )
    with pytest.warns(TypeConversionWarning, match=warning):
        transformed = datetime.transform(dates)
    assert str(transformed.dtype) == "datetime64[ns]"
    assert transformed[2] is pd.NaT
    assert datetime.datetime_format == "%m/%d/%Y"


def test_ordinal_transform(sample_series):
    series_type = str(type(sample_series))
    dask = "dask" in series_type
    spark = "spark" in series_type

    if dask or spark:
        pytest.xfail(
            "Fails with Dask and Spark - ordinal data validation not supported",
        )

    ordinal_incomplete_order = Ordinal(order=["a", "b"])
    error_msg = re.escape(
        "Ordinal column sample_series contains values that are not "
        "present in the order values provided: ['c']",
    )

    with pytest.raises(ValueError, match=error_msg):
        ordinal_incomplete_order.transform(sample_series)


def test_ordinal_validate(sample_series):
    series_type = str(type(sample_series))
    dask = "dask" in series_type
    spark = "spark" in series_type

    if dask or spark:
        pytest.xfail(
            "Fails with Dask and Spark - ordinal data validation not supported",
        )

    ordinal_incomplete_order = Ordinal(order=["a", "b"])
    error_msg = re.escape(
        "Ordinal column sample_series contains values that are not "
        "present in the order values provided: ['c']",
    )

    with pytest.raises(ValueError, match=error_msg):
        ordinal_incomplete_order.validate(sample_series)

    new_type = "string"
    error_message = re.escape(
        f"Series dtype '{new_type}' is incompatible with ordinal dtype.",
    )
    with pytest.raises(TypeValidationError, match=error_message):
        ordinal_incomplete_order.validate(sample_series.astype(new_type))


def test_email_address_validate(sample_df):
    email_address = EmailAddress()
    dtype = email_address.primary_dtype
    series = sample_df["email"].astype(dtype)
    invalid_row = pd.Series({4: "bad_email"}, name="email").astype(dtype)

    if _is_spark_series(series):
        invalid_row = ps.from_pandas(invalid_row)

    assert email_address.validate(series) is None

    series = concat_dataframe_or_series(series, invalid_row).astype(dtype)

    match = "Series email contains invalid email address values. "
    match += "The email_inference_regex can be changed in the config if needed."

    with pytest.raises(TypeValidationError, match=match):
        email_address.validate(series)

    actual = email_address.validate(series, return_invalid_values=True)
    expected = pd.Series({4: "bad_email"}, name="email").astype(dtype)
    assert to_pandas(actual).equals(expected)


def test_url_validate(sample_df):
    logical_type = URL()
    dtype = logical_type.primary_dtype
    series = sample_df["url"].astype(dtype)
    invalid_row = pd.Series({4: "bad_url"}, name="url").astype(dtype)
    if _is_spark_series(series):
        invalid_row = ps.from_pandas(invalid_row)

    assert logical_type.validate(series) is None

    series = concat_dataframe_or_series(series, invalid_row).astype(dtype)

    match = "Series url contains invalid url values. "
    match += "The url_inference_regex can be changed in the config if needed."

    with pytest.raises(TypeValidationError, match=match):
        logical_type.validate(series)

    actual = logical_type.validate(series, return_invalid_values=True)
    expected = pd.Series({4: "bad_url"}, name="url").astype(dtype)
    assert to_pandas(actual).equals(expected)


@pytest.mark.parametrize(
    argnames="logical_type",
    ids=["age", "age_fractional", "age_nullable"],
    argvalues=[Age(), AgeFractional(), AgeNullable()],
)
def test_age_validate(sample_df, logical_type):
    series = sample_df["age"]
    if isinstance(logical_type, Age):
        series = series.dropna()

    dtype = logical_type.primary_dtype
    series = series.astype(dtype)

    assert logical_type.validate(series, return_invalid_values=False) is None
    invalid_row = pd.Series({4: -3}, name="age", dtype=dtype)

    if _is_spark_series(series):
        invalid_row = ps.from_pandas(invalid_row)

    series = concat_dataframe_or_series(series, invalid_row).astype(dtype)

    match = "Series age contains negative values."
    with pytest.raises(TypeValidationError, match=match):
        logical_type.validate(series, return_invalid_values=False)

    actual = logical_type.validate(series, return_invalid_values=True)
    assert to_pandas(actual).equals(to_pandas(invalid_row))


def test_phone_number_validate(sample_df):
    phone_number = PhoneNumber()
    dtype = phone_number.primary_dtype
    series = sample_df["phone_number"].astype(dtype)
    invalid_row = pd.Series({4: "bad_phone"}, name="phone_number").astype(dtype)

    if _is_spark_series(series):
        invalid_row = ps.from_pandas(invalid_row)

    assert phone_number.validate(series) is None

    series = concat_dataframe_or_series(series, invalid_row).astype(dtype)

    match = "Series phone_number contains invalid phone number values. "
    match += "The phone_inference_regex can be changed in the config if needed."

    with pytest.raises(TypeValidationError, match=match):
        phone_number.validate(series)

    actual = phone_number.validate(series, return_invalid_values=True)
    expected = pd.Series({4: "bad_phone"}, name="phone_number").astype(dtype)
    assert to_pandas(actual).equals(expected)


def test_phone_number_validate_complex(sample_df_phone_numbers):
    phone_number = PhoneNumber()
    dtype = phone_number.primary_dtype
    series = sample_df_phone_numbers["phone_number"].astype(dtype)
    # Current inference function does not match lack of area code
    invalid_row = pd.Series(
        {17: "252 9384", 18: "+1 194 129 1991", 19: "+01 236 248 8482"},
        name="phone_number",
    ).astype(dtype)

    series = concat_dataframe_or_series(series, invalid_row).astype(dtype)

    actual = phone_number.validate(series, return_invalid_values=True)
    expected = pd.Series(
        {17: "252 9384", 18: "+1 194 129 1991", 19: "+01 236 248 8482"},
        name="phone_number",
    ).astype(dtype)
    assert to_pandas(actual).equals(expected)


def test_postal_code_validate(sample_df_postal_code):
    pc = PostalCode()
    series = sample_df_postal_code["postal_code"]
    invalid_types = pd.Series(
        [
            "hello",
            "HELLO",
            "51342-HEL0",
        ],
    )

    series = concat_dataframe_or_series(series, invalid_types)

    series.name = "postal_code"
    match = "Series postal_code contains invalid postal code values. "
    match += "The postal_code_inference_regex can be changed in the config if needed."

    with pytest.raises(TypeValidationError, match=match):
        pc.validate(series)


def test_postal_code_validate_complex(sample_df_postal_code):
    pc = PostalCode()
    series = sample_df_postal_code["postal_code"]
    invalid_types = pd.Series(
        [
            "1234",
            "123455",
            "123456789",
            "1234-65433",
            "K1A0B1",  # Canadian formatting
            "K1A 0B1",
            "K1A-0B1",
            "DT1 1AE",  # UK formatting
            "DT1-1AE",
            "DT11AE",
        ],
    )
    actual = pc.validate(series, return_invalid_values=True)
    assert not len(actual)

    series = concat_dataframe_or_series(series, invalid_types)

    actual = pc.validate(series, return_invalid_values=True)
    pd.testing.assert_series_equal(actual, invalid_types)


def test_postal_code_validate_numeric(postal_code_numeric_series):
    series = init_series(postal_code_numeric_series, logical_type=PostalCode())
    actual = to_pandas(series.ww.validate_logical_type(return_invalid_values=True))
    expected = pd.Series({5: "1234567890"})

    pd.testing.assert_series_equal(
        actual,
        expected,
        check_dtype=False,
        check_categorical=False,
    )


def test_postal_code_error(postal_code_numeric_series_pandas):
    series = concat_dataframe_or_series(
        postal_code_numeric_series_pandas,
        pd.Series([1234.5]),
    )
    match = (
        "Error converting datatype for None from type float64 to type string. "
        "Please confirm the underlying data is consistent with logical type PostalCode."
    )
    with pytest.raises(TypeConversionError, match=match):
        init_series(series, logical_type=PostalCode())


def test_null_invalid_values_double():
    types = {"double": "Double"}
    invalid = "text", None, True, object
    df = pd.DataFrame({"double": [1.2, 3, "4", *invalid]})

    with pytest.raises(
        TypeConversionError,
        match="Please confirm the underlying data is consistent with logical type Double",
    ):
        df.ww.init(logical_types=types, null_invalid_values=False)

    nulls = [None] * len(invalid)
    expected = pd.DataFrame({"double": [1.2, 3.0, 4.0, *nulls]})
    df.ww.init(logical_types=types, null_invalid_values=True)
    pd.testing.assert_frame_equal(df, expected)


def test_null_invalid_values_boolean():
    types = {"data": "BooleanNullable"}
    invalid = "text", 1.2, 345, object, None
    data = [True, "False", *invalid]
    df = pd.DataFrame({"data": data})

    with pytest.raises(
        TypeConversionError,
        match="Please confirm the underlying data is consistent with logical type BooleanNullable",
    ):
        df.ww.init(logical_types=types, null_invalid_values=False)

    nulls = [None] * len(invalid)
    data = [True, False, *nulls]
    expected = pd.DataFrame({"data": pd.Series(data, dtype="boolean")})
    df.ww.init(logical_types=types, null_invalid_values=True)
    pd.testing.assert_frame_equal(df, expected)


def test_null_invalid_values_integer():
    types = {"data": "IntegerNullable"}
    invalid = "text", 6.7, object, None
    data = [1.0, "2", 345, *invalid]
    df = pd.DataFrame({"data": data})

    with pytest.raises(
        TypeConversionError,
        match="Please confirm the underlying data is consistent with logical type IntegerNullable",
    ):
        df.ww.init(logical_types=types, null_invalid_values=False)

    nulls = [None] * len(invalid)
    data = [1, 2, 345, *nulls]
    expected = pd.DataFrame({"data": pd.Series(data, dtype="Int64")})
    df.ww.init(logical_types=types, null_invalid_values=True)
    pd.testing.assert_frame_equal(df, expected)


def test_null_invalid_values_emails():
    types = {"data": "EmailAddress"}
    invalid = ["text", 6.7, object, None]
    valid = ["john.smith@example.com", "support@example.com", "team@example.com"]
    data = pd.Series(valid + invalid)
    df = pd.DataFrame({"data": data})

    df.ww.init(logical_types=types, null_invalid_values=False)
    expected = pd.DataFrame({"data": data.astype("string")})
    pd.testing.assert_frame_equal(df, expected)

    nulls = [None] * len(invalid)
    data = pd.Series(valid + nulls, dtype="string")
    expected = pd.DataFrame({"data": data})
    df.ww.init(logical_types=types, null_invalid_values=True)
    pd.testing.assert_frame_equal(df, expected)


def test_null_invalid_values_url():
    valid = [
        "https://github.com/alteryx",
        "https://twitter.com",
        "http://google.com",
    ]

    types = {"data": "URL"}
    invalid = ["text", 6.7, object, None]
    data = pd.Series(valid + invalid)
    df = pd.DataFrame({"data": data})

    df.ww.init(logical_types=types, null_invalid_values=False)
    expected = pd.DataFrame({"data": data.astype("string")})
    pd.testing.assert_frame_equal(df, expected)

    nulls = [None] * len(invalid)
    data = pd.Series(valid + nulls, dtype="string")
    expected = pd.DataFrame({"data": data})
    df.ww.init(logical_types=types, null_invalid_values=True)
    pd.testing.assert_frame_equal(df, expected)


def test_null_invalid_values_phone_number():
    valid = [
        "2002007865",
        "311-311-3156",
        "422.422.0461",
    ]

    types = {"data": "PhoneNumber"}
    invalid = ["text", 6.7, object, None]
    data = pd.Series(valid + invalid)
    df = pd.DataFrame({"data": data})

    df.ww.init(logical_types=types, null_invalid_values=False)
    expected = pd.DataFrame({"data": data.astype("string")})
    pd.testing.assert_frame_equal(df, expected)

    nulls = [None] * len(invalid)
    data = pd.Series(valid + nulls, dtype="string")
    expected = pd.DataFrame({"data": data})
    df.ww.init(logical_types=types, null_invalid_values=True)
    pd.testing.assert_frame_equal(df, expected)


def test_null_invalid_age_fractional():
    types = {"data": "AgeFractional"}
    invalid = "text", -6.7, object, None
    data = [0.34, "24.34", 45.0, *invalid]
    df = pd.DataFrame({"data": data})

    with pytest.raises(
        TypeConversionError,
        match="Please confirm the underlying data is consistent with logical type AgeFractional",
    ):
        df.ww.init(logical_types=types, null_invalid_values=False)

    nulls = [None] * len(invalid)
    data = [0.34, 24.34, 45.0, *nulls]
    expected = pd.DataFrame({"data": data})
    df.ww.init(logical_types=types, null_invalid_values=True)
    pd.testing.assert_frame_equal(df, expected)


def test_null_invalid_age_nullable():
    types = {"data": "AgeNullable"}
    invalid = "text", -6, object, None
    data = [34, "24", 45, *invalid]
    df = pd.DataFrame({"data": data})

    with pytest.raises(
        TypeConversionError,
        match="Please confirm the underlying data is consistent with logical type AgeNullable",
    ):
        df.ww.init(logical_types=types, null_invalid_values=False)

    nulls = [None] * len(invalid)
    data = pd.Series([34, 24, 45, *nulls], dtype="Int64")
    expected = pd.DataFrame({"data": data})
    df.ww.init(logical_types=types, null_invalid_values=True)
    pd.testing.assert_frame_equal(df, expected)


def test_null_invalid_latlong():
    valid = [
        (33.670914, -117.841501),
        "(40.423599, -86.921162)",
        (-45.031705, None),
        (None, None),
    ]
    types = {"data": "LatLong"}
    invalid = ["text", -6.7, object, None]
    df = pd.DataFrame({"data": valid + invalid})

    with pytest.raises(
        TypeValidationError,
        match="LatLong value is not properly formatted.",
    ):
        df.ww.init(logical_types=types, null_invalid_values=False)

    nan = float("nan")
    nulls = [nan] * len(invalid)
    data = pd.Series(
        [
            (33.670914, -117.841501),
            (40.423599, -86.921162),
            (-45.031705, nan),
            (nan, nan),
            *nulls,
        ],
    )
    expected = pd.DataFrame({"data": data})
    df.ww.init(logical_types=types, null_invalid_values=True)
    pd.testing.assert_frame_equal(df, expected)


def test_null_invalid_postal_code():
    types = {"data": "PostalCode"}
    invalid = ["text", 6.7, object, "123456"]
    valid = [90210, "60018-0123", "10010"]
    data = pd.Series(valid + invalid)
    df = pd.DataFrame({"data": data})

    df.ww.init(logical_types=types, null_invalid_values=False)
    expected = pd.DataFrame({"data": data.astype("category")})
    pd.testing.assert_frame_equal(df, expected)

    nulls = [None] * len(invalid)
    data = pd.Series(valid + nulls, dtype="string")
    expected = pd.DataFrame({"data": data.astype("category")})
    df.ww.init(logical_types=types, null_invalid_values=True)
    pd.testing.assert_frame_equal(df, expected)


def test_null_invalid_postal_code_numeric():
    types = {"data": "PostalCode"}
    invalid = [-6.7, 60018.0123, 123456.0]
    valid = [90210, 60018.0, 10010.0]
    data = pd.Series(valid + invalid)
    df = pd.DataFrame({"data": data})

    with pytest.raises(
        TypeConversionError,
        match="Please confirm the underlying data is consistent with logical type PostalCode",
    ):
        df.ww.init(logical_types=types, null_invalid_values=False)

    nulls = [None] * len(invalid)
    data = pd.Series(valid + nulls, dtype="Int64")
    data = data.astype("string").astype("category")
    expected = pd.DataFrame({"data": data})
    df.ww.init(logical_types=types, null_invalid_values=True)
    pd.testing.assert_frame_equal(df, expected)


@pytest.mark.parametrize(
    "null_type",
    [None, pd.NA, pd.NaT, np.nan, "null", "N/A", "mix", 5],
)
@pytest.mark.parametrize("data_type", [int, float])
def test_integer_nullable(data_type, null_type):
    nullable_nums = pd.DataFrame(
        map(data_type, [1, 2, 3, 4, 5] * 20),
        columns=["num_nulls"],
    )
    nullable_nums["num_nulls"].iloc[-5:] = (
        [None, pd.NA, np.nan, "NA", "none"]
        if not isinstance(null_type, pd._libs.missing.NAType) and null_type == "mix"
        else [null_type] * 5
    )
    nullable_nums.ww.init()

    if not isinstance(null_type, int):
        assert isinstance(nullable_nums.ww.logical_types["num_nulls"], IntegerNullable)
        assert all(nullable_nums["num_nulls"][-5:].isna())
    elif data_type is int:
        assert isinstance(nullable_nums.ww.logical_types["num_nulls"], Integer)
    else:
        assert isinstance(nullable_nums.ww.logical_types["num_nulls"], Double)


@pytest.mark.parametrize(
    "null_type",
    [None, pd.NA, pd.NaT, np.nan, "null", "N/A", "mix", True],
)
def test_boolean_nullable(null_type):
    nullable_bools = pd.DataFrame([True, False] * 50, columns=["bool_nulls"])
    nullable_bools["bool_nulls"].iloc[-5:] = (
        [None, pd.NA, np.nan, "NA", "none"]
        if not isinstance(null_type, pd._libs.missing.NAType) and null_type == "mix"
        else [null_type] * 5
    )
    nullable_bools.ww.init()

    if not isinstance(null_type, bool):
        assert isinstance(
            nullable_bools.ww.logical_types["bool_nulls"],
            BooleanNullable,
        )
        assert all(nullable_bools["bool_nulls"][-5:].isna())
    else:
        assert isinstance(nullable_bools.ww.logical_types["bool_nulls"], Boolean)


def test_replace_nans_same_types():
    series = pd.Series([1, 3, 5, -6, "nan"], dtype="object")

    new_series = _replace_nans(series, LatLong.primary_dtype)  # Object dtype

    assert new_series.dtype == "object"
    pd.testing.assert_series_equal(new_series, pd.Series([1, 3, 5, -6, "nan"]))

    new_series = _replace_nans(series, Double.primary_dtype)

    assert new_series.dtype == "float"
    pd.testing.assert_series_equal(new_series, pd.Series([1.0, 3.0, 5.0, -6.0, np.nan]))


def get_expected_dates(dates):
    expected = []
    for d in dates:
        if d is not None:
            year = int(re.findall(r"\d+", d)[0])  # gets the year
            if year > datetime.today().year + 10:
                year -= 100
            if year <= datetime.today().year - 90:
                year += 100
            expected.append("{}-01-01".format(year))
        else:
            expected.append(d)
    return expected


@pytest.mark.parametrize("delim", ["/", "-", "."])
@pytest.mark.parametrize("dtype", ["string", "object"])
def test_datetime_pivot_point(dtype, delim):
    if dtype == "string" and delim != "/":
        pytest.skip("skipping because we don't want to overtest")
    dates = [
        "01/01/24",
        "01/01/30",
        "01/01/32",
        "01/01/36",
        "01/01/52",
        "01/01/56",
        "01/01/60",
        "01/01/72",
        "01/01/76",
        "01/01/80",
        None,
        "01/01/88",
    ]
    datetime_str = "%m/%d/%y"
    if delim != "/":
        dates = [s.replace("/", delim) if s is not None else s for s in dates]
        datetime_str = datetime_str.replace("/", delim)
    expected_values = [
        "2024-01-01",
        "2030-01-01",
        "2032-01-01",
        "2036-01-01",
        "2052-01-01",
        "2056-01-01",
        "1960-01-01",
        "1972-01-01",
        "1976-01-01",
        "1980-01-01",
        None,
        "1988-01-01",
    ]
    expected_values = get_expected_dates(expected_values)
    df = pd.DataFrame({"dates": dates}, dtype=dtype)
    df_expected = pd.DataFrame({"dates": expected_values}, dtype="datetime64[ns]")
    df.ww.init(logical_types={"dates": Datetime(datetime_format=datetime_str)})
    pd.testing.assert_frame_equal(df, df_expected)


@pytest.mark.parametrize("delim", ["/", "-", ".", ""])
def test_datetime_pivot_point_should_not_apply(delim):
    dates = [
        "01/01/1924",
        "01/01/1928",
        "01/01/1960",
        "01/01/1964",
        "01/01/1968",
        "01/01/1972",
        "01/01/2076",
        "01/01/2088",
    ]
    datetime_str = "%m/%d/%Y"
    if delim == "-":
        dates = [s.replace("/", delim) for s in dates]
        datetime_str = datetime_str.replace("/", delim)
    expected_values = [
        "1924-01-01",
        "1928-01-01",
        "1960-01-01",
        "1964-01-01",
        "1968-01-01",
        "1972-01-01",
        "2076-01-01",
        "2088-01-01",
    ]
    df = pd.DataFrame({"dates": dates})
    df_expected = pd.DataFrame({"dates": expected_values}, dtype="datetime64[ns]")
    df.ww.init(logical_types={"dates": Datetime(datetime_format=datetime_str)})
    pd.testing.assert_frame_equal(df, df_expected)


@pytest.mark.parametrize("type", ["pyspark", "dask"])
def test_pyspark_dask_series(type):
    dates = [
        "01/01/24",
        "01/01/28",
        "01/01/30",
        "01/01/32",
        "01/01/36",
        "01/01/40",
        "01/01/72",
        None,
        "01/01/88",
    ]
    datetime_str = "%m/%d/%y"
    expected_values = [
        "2024-01-01",
        "2028-01-01",
        "2030-01-01",
        "1932-01-01",
        "1936-01-01",
        "1940-01-01",
        "1972-01-01",
        None,
        "1988-01-01",
    ]
    expected_values = get_expected_dates(expected_values)
    df = pd.DataFrame({"dates": dates})
    if type == "pyspark":
        ps = pytest.importorskip(
            "pyspark.pandas",
            reason="Pyspark pandas not installed, skipping",
        )
        df = ps.from_pandas(df)
    else:
        dd = pytest.importorskip(
            "dask.dataframe",
            reason="Dask not installed, skipping",
        )
        df = dd.from_pandas(df, npartitions=2)
    df.ww.init(logical_types={"dates": Datetime(datetime_format=datetime_str)})
    df_expected = pd.DataFrame({"dates": expected_values}, dtype="datetime64[ns]")
    df = to_pandas(df)
    df.sort_index(inplace=True)
    pd.testing.assert_frame_equal(df, df_expected)


def test_datetime_pivot_point_no_format_provided():
    dates = [
        "01/01/24",
        "01/01/30",
        "01/01/32",
        "01/01/36",
        "01/01/52",
        "01/01/56",
        "01/01/60",
        "01/01/72",
        "01/01/76",
        "01/01/80",
        None,
        "01/01/88",
    ]
    expected_values = [
        "2024-01-01",
        "2030-01-01",
        "2032-01-01",
        "2036-01-01",
        "2052-01-01",
        "2056-01-01",
        "2060-01-01",
        "1972-01-01",
        "1976-01-01",
        "1980-01-01",
        None,
        "1988-01-01",
    ]
    df = pd.DataFrame({"dates": dates})
    df_expected = pd.DataFrame({"dates": expected_values}, dtype="datetime64[ns]")
    df.ww.init(logical_types={"dates": Datetime})
    pd.testing.assert_frame_equal(df, df_expected)



import warnings
from inspect import isclass

import woodwork as ww
from woodwork.exceptions import DuplicateTagsWarning, StandardTagsChangedWarning
from woodwork.logical_types import (
    Boolean,
    BooleanNullable,
    Datetime,
    LatLong,
    Ordinal,
    Unknown,
)
from woodwork.utils import _convert_input_to_set


class ColumnSchema(object):
    def __init__(
        self,
        logical_type=None,
        semantic_tags=None,
        use_standard_tags=False,
        description=None,
        origin=None,
        metadata=None,
        validate=True,
    ):
        """Create ColumnSchema

        Args:
            logical_type (LogicalType, optional): The column's LogicalType.
            semantic_tags (str, list, set, optional): The semantic tag(s) specified for the column.
            use_standard_tags (boolean, optional): If True, will add standard semantic tags to the column based
                    on the specified logical type if a logical type is defined for the column. Defaults to False.
            description (str, optional): User description of the column.
            origin (str, optional): Origin of the column (i.e. "base" or "engineered").
            metadata (dict[str -> json serializable], optional): Extra metadata provided by the user. The dictionary must contain
                data types that are JSON serializable such as string, integers, and floats. DataFrame and Series types are not supported.
            validate (bool, optional): Whether to perform parameter validation. Defaults to True.
        """
        metadata = metadata or {}

        if isclass(logical_type):
            logical_type = logical_type()

        if validate:
            if logical_type is not None:
                _validate_logical_type(logical_type)
            _validate_description(description)
            _validate_origin(origin)
            _validate_metadata(metadata)
        self._metadata = metadata
        self._description = description
        self._origin = origin
        self.logical_type = logical_type

        self.use_standard_tags = use_standard_tags

        semantic_tags = self._get_column_tags(semantic_tags, validate)
        self.semantic_tags = semantic_tags

    def __eq__(self, other, deep=True):
        if self.use_standard_tags != other.use_standard_tags:
            return False
        if self.logical_type != other.logical_type:
            return False
        if self.semantic_tags != other.semantic_tags:
            return False
        if self.description != other.description:
            return False
        if self.origin != other.origin:
            return False
        if deep and self.metadata != other.metadata:
            return False

        return True

    def __repr__(self):
        msg = "<ColumnSchema"
        if self.logical_type is not None:
            msg += " (Logical Type = {})".format(self.logical_type)
        if self.semantic_tags:
            msg += " (Semantic Tags = {})".format(sorted(list(self.semantic_tags)))
        msg += ">"
        return msg

    def _get_column_tags(self, semantic_tags, validate):
        semantic_tags = _convert_input_to_set(
            semantic_tags,
            error_language="semantic_tags",
            validate=validate,
        )

        if self.use_standard_tags:
            if self.logical_type is None:
                raise ValueError("Cannot use standard tags when logical_type is None")
            semantic_tags = semantic_tags.union(self.logical_type.standard_tags)

        return semantic_tags

    @property
    def description(self):
        """Description of the column"""
        return self._description

    @description.setter
    def description(self, description):
        _validate_description(description)
        self._description = description

    @property
    def origin(self):
        """Origin of the column"""
        return self._origin

    @origin.setter
    def origin(self, origin):
        _validate_origin(origin)
        self._origin = origin

    @property
    def metadata(self):
        """Metadata of the column"""
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        metadata = metadata or {}
        _validate_metadata(metadata)
        self._metadata = metadata

    @property
    def is_numeric(self):
        """Whether the ColumnSchema is numeric in nature"""
        return (
            self.logical_type is not None
            and "numeric" in self.logical_type.standard_tags
        )

    @property
    def is_categorical(self):
        """Whether the ColumnSchema is categorical in nature"""
        return (
            self.logical_type is not None
            and "category" in self.logical_type.standard_tags
        )

    @property
    def is_datetime(self):
        """Whether the ColumnSchema is a Datetime column"""
        return type(self.logical_type) == Datetime

    @property
    def is_latlong(self):
        """Whether the ColumnSchema is a LatLong column"""
        return type(self.logical_type) == LatLong

    @property
    def is_boolean(self):
        """Whether the ColumnSchema is a Boolean column"""
        ltype_class = type(self.logical_type)
        return ltype_class == Boolean or ltype_class == BooleanNullable

    @property
    def is_unknown(self):
        """Whether the ColumnSchema is a Unknown column"""
        return type(self.logical_type) == Unknown

    @property
    def is_ordinal(self):
        """Whether the ColumnSchema is a Ordinal column"""
        return type(self.logical_type) == Ordinal

    def _add_semantic_tags(self, new_tags, name):
        """Add the specified semantic tags to the current set of tags

        Args:
            new_tags (str/list/set): The new tags to add
            name (str): Name of the column to use in warning
        """
        new_tags = _convert_input_to_set(new_tags)

        duplicate_tags = sorted(list(self.semantic_tags.intersection(new_tags)))
        if duplicate_tags:
            warnings.warn(
                DuplicateTagsWarning().get_warning_message(duplicate_tags, name),
                DuplicateTagsWarning,
            )
        self.semantic_tags = self.semantic_tags.union(new_tags)

    def _remove_semantic_tags(self, tags_to_remove, name):
        """Removes specified semantic tags from from the current set of tags

        Args:
            tags_to_remove (str/list/set): The tags to remove
            name (str): Name of the column to use in warning
        """
        tags_to_remove = _convert_input_to_set(tags_to_remove)
        invalid_tags = sorted(list(tags_to_remove.difference(self.semantic_tags)))
        if invalid_tags:
            raise LookupError(
                f"Semantic tag(s) '{', '.join(invalid_tags)}' not present on column '{name}'",
            )

        if self.use_standard_tags and sorted(
            list(tags_to_remove.intersection(self.logical_type.standard_tags)),
        ):
            warnings.warn(
                StandardTagsChangedWarning().get_warning_message(
                    not self.use_standard_tags,
                    name,
                ),
                StandardTagsChangedWarning,
            )
        self.semantic_tags = self.semantic_tags.difference(tags_to_remove)

    def _reset_semantic_tags(self):
        """Reset the set of semantic tags to the default values. The default values
        will be either an empty set or the standard tags, controlled by the
        use_standard_tags boolean.
        """
        new_tags = set()
        if self.use_standard_tags:
            new_tags = set(self.logical_type.standard_tags)
        self.semantic_tags = new_tags

    def _set_semantic_tags(self, semantic_tags):
        """Replace current semantic tags with new values. If use_standard_tags is set
        to True, standard tags will be added as well.

        Args:
            semantic_tags (str/list/set): New semantic tag(s) to set
        """
        semantic_tags = _convert_input_to_set(semantic_tags)

        if self.use_standard_tags:
            semantic_tags = semantic_tags.union(self.logical_type.standard_tags)

        self.semantic_tags = semantic_tags

    @property
    def custom_tags(self):
        """The custom semantic tag(s) specified for the column."""
        standard_tags = set()
        if self.use_standard_tags:
            standard_tags |= self.logical_type.standard_tags
        return self.semantic_tags - standard_tags - {"index", "time_index"}


def _validate_logical_type(logical_type):
    if type(logical_type) not in ww.type_system.registered_types:
        raise TypeError(f"logical_type {logical_type} is not a registered LogicalType.")


def _validate_description(column_description):
    if column_description is not None and not isinstance(column_description, str):
        raise TypeError("Column description must be a string")


def _validate_origin(origin):
    if origin is not None and not isinstance(origin, str):
        raise TypeError("Column origin must be a string")


def _validate_metadata(column_metadata):
    if not isinstance(column_metadata, dict):
        raise TypeError("Column metadata must be a dictionary")



def _get_top_values_categorical(series, num_x):
    """Get the most frequent values in a pandas Series. Will exclude null values.

    Args:
        column (pd.Series): data to use find most frequent values
        num_x (int): the number of top values to retrieve

    Returns:
        top_list (list(dict)): a list of dictionary with keys `value` and `count`.
            Output is sorted in descending order based on the value counts.
    """
    frequencies = series.value_counts(dropna=True)
    df = frequencies.head(num_x).reset_index()
    df.columns = ["value", "count"]
    df = df.sort_values(["count", "value"], ascending=[False, True])
    value_counts = list(df.to_dict(orient="index").values())
    return value_counts
